@@ -20,6 +20,7 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
  *
  */
+#define _GNU_SOURCE
 #include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -29,59 +30,70 @@
 #include <assert.h>
 #include <string.h>
 #include <err.h>
+#include <errno.h>
 
 
 #include "gpg-common.h"
 #include "multiplex.h"
 
+static int validate_fd_argument(const char *const untrusted_fd_arg) {
+    if (untrusted_fd_arg == NULL || !untrusted_fd_arg[0])
+        goto fail;
+    const char *untrusted_p = untrusted_fd_arg;
+    for (; *untrusted_p; untrusted_p++)
+        if (*untrusted_p < '0' || *untrusted_p > '9')
+            goto fail;
+    if (untrusted_fd_arg[0] <= '0' && untrusted_fd_arg[1]) {
+        fprintf(stderr, "Leading zeroes in FD argument %s not allowed\n", untrusted_fd_arg);
+        exit(1);
+    }
+    if (untrusted_p - untrusted_fd_arg > 4)
+        goto too_big;
+    errno = 0;
+    char *endptr = NULL;
+    long const untrusted_fd = strtol(untrusted_fd_arg, &endptr, 10);
+    if (untrusted_fd < 0 || untrusted_fd > 9999 || errno || !endptr || *endptr)
+        abort(); // should have been caught earlier
+    if (untrusted_fd >= MAX_FD_VALUE) {
+too_big:
+        fprintf(stderr, "FD value too big (%s > %d)\n",
+                untrusted_fd_arg, MAX_FD_VALUE - 1);
+        exit(1);
+    }
+    return (int)untrusted_fd;
+fail:
+    fprintf(stderr, "Invalid fd argument.  Only decimal numbers are supported.\n");
+    exit(1);
+}
+
 /* Add current argument (optarg) to given list
  * Check for its correctness
  */
 void add_arg_to_fd_list(int *list, int *list_count)
 {
-    int i;
-    char *endptr;
-    int cur_fd;
-    long untrusted_cur_fd;
+    int i, cur_fd, untrusted_cur_fd;
 
     if (*list_count >= MAX_FDS - 1) {
         fprintf(stderr, "Too many FDs specified\n");
         exit(1);
     }
-    /* optarg is untrusted! */
-    if (optarg == NULL || optarg[0] < '0' || optarg[0] > '9')
-        goto fail;
-    untrusted_cur_fd = strtol(optarg, &endptr, 0);
-    if (untrusted_cur_fd < 0)
-        abort(); // should have been caught by the optarg[0] checks
-    if (endptr != NULL && endptr[0] == 0) {
-        // limit fd value
-        if (untrusted_cur_fd > MAX_FD_VALUE) {
-            fprintf(stderr, "FD value too big (%ld > %d)\n",
-                    untrusted_cur_fd, MAX_FD_VALUE);
-            exit(1);
-        }
-        // check if not already in list
-        for (i = 0; i < *list_count; i++) {
-            if (list[i] == (int)untrusted_cur_fd)
-                break;
-        }
-        cur_fd = (int)untrusted_cur_fd;
-        /* FD sanitization end */
-        if (i == *list_count)
-            list[(*list_count)++] = cur_fd;
-    } else {
-fail:
-        fprintf(stderr, "Invalid fd argument\n");
-        exit(1);
+    untrusted_cur_fd = validate_fd_argument(optarg);
+    // check if not already in list
+    for (i = 0; i < *list_count; i++) {
+        if (list[i] == (int)untrusted_cur_fd)
+            break;
     }
+    cur_fd = untrusted_cur_fd;
+    /* FD sanitization end */
+    if (i == *list_count)
+        list[(*list_count)++] = cur_fd;
 }
 
 void handle_opt_verify(char *untrusted_sig_path, int *list, int *list_count, int is_client)
 {
     int i;
     char *sig_path;
-    int cur_fd, untrusted_cur_fd;
+    int cur_fd;
     int untrusted_sig_path_len;
     int fd_path_len;
 
@@ -93,19 +105,8 @@ void handle_opt_verify(char *untrusted_sig_path, int *list, int *list_count, int
         fprintf(stderr, "Invalid fd argument\n");
         exit(1);
     }
-    if (sscanf(untrusted_sig_path, "/dev/fd/%d", &untrusted_cur_fd) > 0) {
-        if (untrusted_cur_fd < 0) {
-            fprintf(stderr, "Invalid fd argument\n");
-            exit(1);
-        }
-        // limit fd value
-        if (untrusted_cur_fd > MAX_FD_VALUE) {
-            fprintf(stderr, "FD value too big (%d > %d)\n",
-                    untrusted_cur_fd, MAX_FD_VALUE);
-            exit(1);
-        }
-        cur_fd = untrusted_cur_fd;
-        /* FD sanitization end */
+    if (!strncmp(untrusted_sig_path, "/dev/fd/", 8)) {
+        cur_fd = validate_fd_argument(untrusted_sig_path + 8);
     } else {
         if (!is_client) {
             fprintf(stderr, "--verify with filename allowed only on the client side\n");
@@ -113,7 +114,7 @@ void handle_opt_verify(char *untrusted_sig_path, int *list, int *list_count, int
         }
         /* arguments on client side are trusted */
         sig_path = untrusted_sig_path;
-        cur_fd = open(sig_path, O_RDONLY);
+        cur_fd = open(sig_path, O_RDONLY|O_CLOEXEC|O_NOCTTY);
         if (cur_fd < 0) {
             perror("open sig");
             exit(1);
@@ -442,36 +443,30 @@ int parse_options(int argc, char *untrusted_argv[], int *input_fds,
     return optind;
 }
 
-void move_fds(int *dest_fds, int count, int (*pipes)[2], int pipe_end)
+void move_fds(const int *const dest_fds, int const count, int (*const pipes)[2],
+              _Bool pipe_end)
 {
-    int remap_fds[MAX_FD_VALUE * 2];
-    int i;
+    _Static_assert(MAX_FDS > 0 && MAX_FDS < MAX_FD_VALUE, "bad constants");
+    assert(count >= 0 && count <= MAX_FDS);
 
-    for (i = 0; i < MAX_FD_VALUE * 2; i++)
-        remap_fds[i] = -1;
-
-    // close the other ends of pipes
-    for (i = 0; i < count; i++)
-        close(pipes[i][!pipe_end]);
 
     // move pipes to correct fds
-    for (i = 0; i < count; i++) {
-        // if it is currently used - move to other fd and save new position in
-        // remap_fds table
-        if (fcntl(dest_fds[i], F_GETFD) >= 0) {
-            remap_fds[dest_fds[i]] = dup(dest_fds[i]);
-            if (remap_fds[dest_fds[i]] < 0) {
-                // no message - stderr closed
-                exit(1);
-            }
-        }
-        // find pipe end - possibly remapped
-        while (remap_fds[pipes[i][pipe_end]] >= 0)
-            pipes[i][pipe_end] = remap_fds[pipes[i][pipe_end]];
-        if (dest_fds[i] != pipes[i][pipe_end]) {
-            // move fd to destination position
-            dup2(pipes[i][pipe_end], dest_fds[i]);
-            close(pipes[i][pipe_end]);
+    for (int i = 0; i < count; i++)
+        if (dup2(pipes[i][pipe_end], dest_fds[i]) != dest_fds[i])
+            _exit(1);
+    // no need to close the pipe fds as they are CLOEXEC and we exec later
+}
+
+static void dup_over_fd(int const fallback_fd, int const fd) {
+    if (fallback_fd < 0 || fd < 0)
+        abort();
+    if (fcntl(fd, F_GETFD) == -1) {
+        assert(errno == EBADF);
+        int const new_fd = fcntl(fallback_fd, F_DUPFD_CLOEXEC, fd);
+        if (new_fd != fd) {
+            assert(new_fd == -1 && "F_DUPFD_CLOEXEC set file descriptor to bad value?");
+            perror("dup2");
+            exit(1);
         }
     }
 }
@@ -490,10 +485,23 @@ int prepare_pipes_and_run(const char *run_file, char **run_argv, int *input_fds,
 
     sigemptyset(&chld_set);
     sigaddset(&chld_set, SIGCHLD);
-
+    if (input_fds_count > MAX_FDS || output_fds_count > MAX_FDS)
+        abort();
+    else {
+        int const null_fd = open("/dev/null", O_RDONLY | O_NOCTTY | O_CLOEXEC | O_NOFOLLOW);
+        if (null_fd == -1) {
+            perror("open /dev/null");
+            exit(1);
+        }
+        for (i = 0; i < input_fds_count; ++i)
+            dup_over_fd(null_fd, input_fds[i]);
+        for (i = 0; i < output_fds_count; ++i)
+            dup_over_fd(null_fd, output_fds[i]);
+        close(null_fd);
+    }
 
     for (i = 0; i < input_fds_count; i++) {
-        if (pipe(pipes_in[i]) < 0) {
+        if (pipe2(pipes_in[i], O_CLOEXEC) < 0) {
             perror("pipe");
             exit(1);
         }
@@ -501,7 +509,7 @@ int prepare_pipes_and_run(const char *run_file, char **run_argv, int *input_fds,
         pipes_in_for_multiplexer[i] = pipes_in[i][1];
     }
     for (i = 0; i < output_fds_count; i++) {
-        if (pipe(pipes_out[i]) < 0) {
+        if (pipe2(pipes_out[i], O_CLOEXEC) < 0) {
             perror("pipe");
             exit(1);
         }
@@ -509,7 +517,7 @@ int prepare_pipes_and_run(const char *run_file, char **run_argv, int *input_fds,
         pipes_out_for_multiplexer[i] = pipes_out[i][0];
     }
 
-    setup_sigchld();
+    setup_sigchld(false);
 
     switch (pid = fork()) {
         case -1:
@@ -517,14 +525,10 @@ int prepare_pipes_and_run(const char *run_file, char **run_argv, int *input_fds,
             exit(1);
         case 0:
             // child
-            close(0);
-            close(1);
-            close(2);
             move_fds(input_fds, input_fds_count, pipes_in, 0);
             move_fds(output_fds, output_fds_count, pipes_out, 1);
             execv(run_file, run_argv);
-            //error message already printed by getopt
-            exit(1);
+            _exit(1);
         default:
             // close unneded end of pipes
             for (i = 0; i < input_fds_count; i++)
